#include "Morpher.h"

void Morpher::interpolate_lines(double t) {
  list<Shape *>::iterator j = targetlines->begin();
  interlines.clear();
  for (list<Shape *>::iterator i = sourcelines->begin(); i != sourcelines->end(); i++) {
    double x1 = (1-t)*((Line *)(*i))->get_x1() + t*((Line *)(*j))->get_x1();
    double y1 = (1-t)*((Line *)(*i))->get_y1() + t*((Line *)(*j))->get_y1();
    double x2 = (1-t)*((Line *)(*i))->get_x2() + t*((Line *)(*j))->get_x2();
    double y2 = (1-t)*((Line *)(*i))->get_y2() + t*((Line *)(*j))->get_y2();
    
    Line *interline = new Line();
    interline->setcoords(x1,y1,x2,y2);
    //    cout <<"("<< x1 <<","<< y1 <<")  -  ("<< x2 <<","<< y2 <<")\n"; 
    interlines.push_back(interline);
    j++;
  }
}

double Morpher::mapu(double x, double y, list<Shape *>::iterator i) {
  double px = ((Line *)(*i))->get_x1();
  double py = ((Line *)(*i))->get_y1();
  double qx = ((Line *)(*i))->get_x2();
  double qy = ((Line *)(*i))->get_y2();
  double u = ((x - px) * (qx - px) + (y - py) * (qy - py)) / (pow(qx-px,2)+pow(qy-py,2));
  //  cout <<"("<< x <<","<< y <<")  ->  u = "<< u <<"\n"; 
  return u;
}

double Morpher::mapv(double x, double y, list<Shape *>::iterator i) {
  double px = ((Line *)(*i))->get_x1();
  double py = ((Line *)(*i))->get_y1();
  double qx = ((Line *)(*i))->get_x2();
  double qy = ((Line *)(*i))->get_y2();
  double v = ((x - px) * (py - qy) + (y - py) * (qx - px)) / sqrt(pow(qx-px,2)+pow(qy-py,2));
  //  cout <<"("<< x <<","<< y <<")  ->  v = "<< v <<"\n"; 
  return v;
}

double Morpher::backmapx(double x, double y, double u, double v, list<Shape *>::iterator i) {
  double px = ((Line *)(*i))->get_x1();
  double py = ((Line *)(*i))->get_y1();
  double qx = ((Line *)(*i))->get_x2();
  double qy = ((Line *)(*i))->get_y2();

  double newx = px + u * (qx - px) + v * (py - qy) / sqrt(pow(qx-px,2)+pow(qy-py,2));
  return newx;
}

double Morpher::backmapy(double x, double y, double u, double v, list<Shape *>::iterator i) {
  double px = ((Line *)(*i))->get_x1();
  double py = ((Line *)(*i))->get_y1();
  double qx = ((Line *)(*i))->get_x2();
  double qy = ((Line *)(*i))->get_y2();

  double newy = py + u * (qy - py) + v * (qx - px) / sqrt(pow(qx-px,2)+pow(qy-py,2));
  return newy;
}

void Morpher::singlemorph(double t) {
  int width = VImageNColumns(sourceim[0]);
  int height = VImageNRows(sourceim[0]);
  msource = VCreateImage(3,height,width,VUByteRepn);
  mtarget = VCreateImage(3,height,width,VUByteRepn);
  morphed = VCreateImage(3,height,width,VUByteRepn);
  interpolate_lines(t);
  for (int x = 0; x < width; x++) {
    for (int y = 0; y < height; y++) {
      list<Shape *>::iterator i = sourcelines->begin();
      list<Shape *>::iterator j = targetlines->begin();
      list<Shape *>::iterator k = interlines.begin();
      double dsumsrc_x = 0;
      double dsumsrc_y = 0;
      double dsumtar_x = 0;
      double dsumtar_y = 0;
      double weightsum = 0;

      for (; i != sourcelines->end();) {
	double u = mapu(x,y,k);
	double v = mapv(x,y,k);
	double src_x = backmapx(x,y,u,v,i);
	double src_y = backmapy(x,y,u,v,i);
	double tar_x = backmapx(x,y,u,v,j);
	double tar_y = backmapy(x,y,u,v,j);
	double dsrc_x = src_x - x;
	double dsrc_y = src_y - y;
	double dtar_x = tar_x - x;
	double dtar_y = tar_y - y;
	double dist;

	double px = ((Line *)(*k))->get_x1();
	double py = ((Line *)(*k))->get_y1();
	double qx = ((Line *)(*k))->get_x2();
	double qy = ((Line *)(*k))->get_y2();
	if (u >= 0 && u <= 1) {
	  dist = fabs(v);
	} else if (u > 1) {
	  dist = sqrt(pow(qx-x,2)+pow(qy-y,2));
	} else {
	  dist = sqrt(pow(px-x,2)+pow(py-y,2));
	}
	
	double length = sqrt(pow(px-qx,2)+pow(py-qy,2));
	double p=1,a=1,b=2;
	double weight = pow((pow(length,p)/(a+dist)),b);
	
	dsumsrc_x += weight*dsrc_x;
	dsumsrc_y += weight*dsrc_y;
	dsumtar_x += weight*dtar_x;
	dsumtar_y += weight*dtar_y;

	weightsum += weight;

	i++;
	j++;
	k++;
      }
      double finalsrc_x = x + dsumsrc_x/weightsum;
      double finalsrc_y = y + dsumsrc_y/weightsum;
      double finaltar_x = x + dsumtar_x/weightsum;
      double finaltar_y = y + dsumtar_y/weightsum;

      if (finalsrc_x >= 0 && finalsrc_x < VImageNColumns(sourceim[0]) && finalsrc_y >= 0 && finalsrc_y < VImageNRows(sourceim[0])) {
	VPixel(msource, 0, y , x, VUByte) = VPixel(sourceim[0], 0, int(finalsrc_y), int(finalsrc_x), VUByte);
	VPixel(msource, 1, y , x, VUByte) = VPixel(sourceim[0], 1, int(finalsrc_y), int(finalsrc_x), VUByte);
	VPixel(msource, 2, y , x, VUByte) = VPixel(sourceim[0], 2, int(finalsrc_y), int(finalsrc_x), VUByte);
      } else {
	VPixel(msource, 0, y , x, VUByte) = 150;
	VPixel(msource, 1, y , x, VUByte) = 110;
	VPixel(msource, 2, y , x, VUByte) = 85;
      }
      if (finaltar_x >= 0 && finaltar_x < VImageNColumns(targetim[0]) && finaltar_y >= 0 && finaltar_y < VImageNRows(targetim[0])) {
      	VPixel(mtarget, 0, y , x, VUByte) = VPixel(targetim[0], 0, int(finaltar_y), int(finaltar_x), VUByte);
      	VPixel(mtarget, 1, y , x, VUByte) = VPixel(targetim[0], 1, int(finaltar_y), int(finaltar_x), VUByte);
      	VPixel(mtarget, 2, y , x, VUByte) = VPixel(targetim[0], 2, int(finaltar_y), int(finaltar_x), VUByte);
      } else {
	VPixel(mtarget, 0, y , x, VUByte) = 150;
	VPixel(mtarget, 1, y , x, VUByte) = 110;
	VPixel(mtarget, 2, y , x, VUByte) = 85;
      }

    }
  }
  
  for (int x = 0; x < width; x++) {
    for (int y = 0; y < height; y++) {
      VPixel(morphed,0,y,x,VUByte) = (VUByte)((1-t)*VPixel(msource,0,y,x,VUByte) + t*VPixel(mtarget,0,y,x,VUByte));
      VPixel(morphed,1,y,x,VUByte) = (VUByte)((1-t)*VPixel(msource,1,y,x,VUByte) + t*VPixel(mtarget,1,y,x,VUByte));
      VPixel(morphed,2,y,x,VUByte) = (VUByte)((1-t)*VPixel(msource,2,y,x,VUByte) + t*VPixel(mtarget,2,y,x,VUByte));
    }
  }
}

void Morpher::multimorph(int nummorphs, char* filename) {
  double morphinc = 1.0 / nummorphs;
  double t = morphinc;
  char curfilename[20];
  int index = 1;
  while (t < 1) {
    singlemorph(t);
    sprintf(curfilename,"%s%d.v",filename,index);
    VWriteImages(fopen(curfilename,"wb"),NULL, 1, &morphed);
    t += morphinc;
    index++;
  }
}

VImage Morpher::getmorphedsource() {
  return msource;
}

VImage Morpher::getmorphedtarget() {
  return mtarget;
}

VImage Morpher::getmorphedimage() {
  return morphed;
}
